"""
--- Day 5: How About a Nice Game of Chess? ---
https://adventofcode.com/2016/day/5
"""
from _md5 import md5

from aocd import data


code1 = []
code2 = ["."] * 8
remaining = set("01234567")
h0 = md5(data.encode("ascii")).copy


n = 0
while True:
    hash_ = h0()
    hash_.update(b"%d" % n)
    hash_ = hash_.hexdigest()
    n += 1
    if not hash_.startswith("00000"):
        continue
    h5, h6 = hash_[5], hash_[6]
    code1.append(h5)
    if h5 in remaining:
        code2[int(h5)] = h6
        remaining.remove(h5)
        if not remaining:
            break

code1 = "".join(code1[:8])
code2 = "".join(code2)


print("answer_a:", code1)
print("answer_b:", code2)



"""
--- Day 6: Chronal Coordinates ---
https://adventofcode.com/2018/day/6
"""
from io import StringIO

import numpy as np
from aocd import data


vs = np.loadtxt(StringIO(data), dtype=int, delimiter=",")
w, h = vs.max(axis=0) + 1
n = len(vs)
d_max = 32 if n == 6 else 10000
ps = np.mgrid[:w, :h].transpose(1, 2, 0).reshape(-1, 1, 2)
ds = np.abs(ps - vs).sum(axis=2)
d = ds.argmin(axis=1)
ties = d != n - 1 - np.fliplr(ds).argmin(axis=1)
d[ties] = -1
border = {*d[:w], *d[-w:], *d[::h], *d[::-h]}
areas = [(d == c).sum() for c in range(n) if c not in border]
print("answer_a:", max(areas))
print("answer_b:", (ds.sum(axis=1) < d_max).sum())



"""
--- Day 6: Tuning Trouble ---
https://adventofcode.com/2022/day/6
"""
from collections import Counter

from aocd import data


def marker(data, n):
    counts = Counter(data[:n])
    i = n
    while len(counts) < n:
        if data[i] != data[i - n]:
            counts += {data[i]: 1, data[i - n]: -1}
        i += 1
    return i


print("answer_a:", marker(data, 4))
print("answer_b:", marker(data, 14))



"""
--- Day 13: Point of Incidence ---
https://adventofcode.com/2023/day/13
"""
import numpy as np
from aocd import data

a = b = 0
for chunk in data.split("\n\n"):
    h = chunk.count("\n") + 1
    A = np.fromiter(chunk.replace("\n", ""), dtype="<U1").reshape(h, -1)
    for A, f in zip([A, A.T], [100, 1]):
        h = len(A)
        for i in range(1, h):
            A1 = A[max(2*i-h,0):i]
            A2 = A[i:i+min(i,h-i)][::-1]
            match (A1 != A2).sum():
                case 0: a += f * i
                case 1: b += f * i

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 21: RPG Simulator 20XX ---
https://adventofcode.com/2015/day/21
"""
from collections import defaultdict
from itertools import combinations

from aocd import data
from fields import Fields
from fields import Tuple


shop_data = """\
Weapons:    Cost  Damage  Armor
Dagger        8     4       0
Shortsword   10     5       0
Warhammer    25     6       0
Longsword    40     7       0
Greataxe     74     8       0

Armor:      Cost  Damage  Armor
Leather      13     0       1
Chainmail    31     0       2
Splintmail   53     0       3
Bandedmail   75     0       4
Platemail   102     0       5

Rings:      Cost  Damage  Armor
Damage +1    25     1       0
Damage +2    50     2       0
Damage +3   100     3       0
Defense +1   20     0       1
Defense +2   40     0       2
Defense +3   80     0       3
"""


class Item(Tuple.name.cost.damage.armor):
    @classmethod
    def fromline(cls, line):
        name, cost, damage, armor = line.rsplit(None, 3)
        item = Item(name, cost=int(cost), damage=int(damage), armor=int(armor))
        return item


shop = defaultdict(list)
for line in shop_data.splitlines():
    if ":" in line:
        itemtype = line.split(":")[0].lower().rstrip("s")
    else:
        if line:
            item = Item.fromline(line)
            shop[itemtype].append(item)


class Player(Fields.damage.armor.name["player"].hp[100]):
    ...


def battle_winner(attacker, defender):
    while attacker.hp > 0 and defender.hp > 0:
        defender.hp -= max(attacker.damage - defender.armor, 1)
        attacker, defender = defender, attacker
    return defender


def parsed(data):
    d = {k.lower(): int(v) for k, v in (line.split(": ") for line in data.splitlines())}
    d["hp"] = d.pop("hit points")
    return Player(name="boss", **d)


def choices(shop):
    weapon_choices = combinations(shop["weapon"], 1)

    armor_choices = [()]  # no armor..
    armor_choices += combinations(shop["armor"], 1)

    ring_choices = [()]  # no rings..
    ring_choices += combinations(shop["ring"], 1)
    ring_choices += combinations(shop["ring"], 2)

    for weapon in weapon_choices:
        for armor in armor_choices:
            for rings in ring_choices:
                items = [*weapon, *armor, *rings]
                total_damage = sum(item.damage for item in items)
                total_armor = sum(item.armor for item in items)
                total_cost = sum(item.cost for item in items)
                player = Player(damage=total_damage, armor=total_armor)
                player.items = items
                player.cost = total_cost
                yield player


def play(data, shop):
    min_cost = float("inf")
    max_cost = 0
    for player in choices(shop):
        boss = parsed(data)
        if battle_winner(player, boss) is player:
            min_cost = min(player.cost, min_cost)
        else:
            max_cost = max(player.cost, max_cost)
    return min_cost, max_cost


if __name__ == "__main__":
    min_cost_to_win, max_cost_to_lose = play(data, shop)
    print("answer_a:", min_cost_to_win)
    print("answer_b:", max_cost_to_lose)



"""
--- Day 8: Treetop Tree House ---
https://adventofcode.com/2022/day/8
"""
from aocd import data

from aoc_wim.zgrid import ZGrid
from aoc_wim.zgrid import zrange

grid = ZGrid(data, transform=int)
h, w = grid.height, grid.width
a = 2*h + 2*w - 4  # perimeter trees are always visible
b = 0
for z0 in zrange(complex(1, 1), complex(w - 1, h - 1)):  # interior trees
    n = grid[z0]
    vis = False  # visibility from exterior
    ss = 1  # scenic score
    for dz in 1, -1, 1j, -1j:
        z = z0
        s = 0
        while True:
            z += dz
            if z not in grid:
                vis = True
                break
            s += 1
            if grid[z] >= n:
                break
        ss *= s
    a += vis
    b = max(b, ss)

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 25: Full of Hot Air ---
https://adventofcode.com/2022/day/25
"""
from aocd import data


def snafu2dec(s):
    result = 0
    for i, char in enumerate(reversed(s)):
        n = -1 if char == "-" else -2 if char == "=" else int(char)
        result += n * (5 ** i)
    return result


def dec2snafu(n):
    result = []
    while n:
        n, r = divmod(n, 5)
        result.append(r)
    for i, val in enumerate(result):
        if val > 2:
            result[i] = "=" if val == 3 else "-" if val == 4 else 0
            try:
                result[i + 1] += 1
            except IndexError:
                result.append(1)
    return "".join(map(str, reversed(result)))


print("answer_a:", dec2snafu(sum(snafu2dec(n) for n in data.split())))



"""
--- Day 8: Haunted Wasteland ---
https://adventofcode.com/2023/day/8
"""
from math import lcm

from aocd import data

path, graph = data.split("\n\n")
d = {}
for line in graph.splitlines():
    S, LR = line.split(" = ")
    d[S] = LR.strip("()").split(", ")


def sync(nodes):
    i = 0
    n = len(nodes)
    z0 = [None] * n
    while True:
        s = path[i % len(path)]
        nodes = [d[x][s == "R"] for x in nodes]
        i += 1
        for j in range(n):
            if nodes[j].endswith("Z") and z0[j] is None:
                z0[j] = i
                if all(z0):
                    return lcm(*z0)


if "AAA" in d:
    print("answer_a:", sync(["AAA"]))
print("answer_b:", sync([x for x in d if x.endswith("A")]))



"""
--- Day 25: Sea Cucumber ---
https://adventofcode.com/2021/day/25
"""
from aocd import data

from aoc_wim.zgrid import ZGrid


def evolve(g):
    moved_right = {}
    for z, zr, _ in z_zr_zd:
        if g[z] == ">" and g[zr] == ".":
            moved_right[z], moved_right[zr] = g[zr], g[z]
    g.update(moved_right)
    moved_down = {}
    for z, _, zd in z_zr_zd:
        if g[z] == "v" and g[zd] == ".":
            moved_down[z], moved_down[zd] = g[zd], g[z]
    g.update(moved_down)
    return len(moved_right) + len(moved_down)


grid = ZGrid(data)
H, W = grid.height, grid.width
z_zr_zd = [
    (complex(x, y), complex((x + 1) % W, y), complex(x, (y + 1) % H))
    for x in range(W)
    for y in range(H)
]
n = 0
while evolve(grid):
    n += 1

print("answer_a:", n + 1)



"""
--- Day 15: Lens Library ---
https://adventofcode.com/2023/day/15
"""
from aocd import data


def keyfunc(s):
    i = 0
    for char in s:
        i += ord(char)
        i *= 17
        i %= 256
    return i


hashmap = [{} for i in range(256)]
for op in data.split(","):
    match op.strip("-").split("="):
        case [k, v]: hashmap[keyfunc(k)][k] = int(v)
        case [k]: hashmap[keyfunc(k)].pop(k, None)

print("answer_a:", sum(keyfunc(op) for op in data.split(",")))
print("answer_b:", sum(i*sum(k*v for k,v in enumerate(box.values(),1)) for i,box in enumerate(hashmap,1)))



"""
--- Day 2: Password Philosophy ---
https://adventofcode.com/2020/day/2
"""
from aocd import data
from parse import parse

a = b = 0
for line in data.splitlines():
    x, y, char, passwd = parse("{:d}-{:d} {}: {}", line)
    a += x <= passwd.count(char) <= y
    b += (passwd[x - 1] + passwd[y - 1]).count(char) == 1

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 20: Pulse Propagation ---
https://adventofcode.com/2023/day/20
"""

from aocd import data
from collections import deque
import logging
import math


# logging.basicConfig(format="%(message)s", level=logging.DEBUG)
log = logging.getLogger(__name__)


pending = deque([])


class Module:
    def __init__(self, pk):
        self.pk = pk
        self.dsts = []
        self.reset()

    def reset(self):
        self._state = 0
        self._n_sent = [0, 0]
        self._recv = []
        self.disabled = False

    @classmethod
    def from_str(cls, s):
        s0, pk = s[:1], s[1:]
        if s0 == "%":
            return FlipFlop(pk)
        elif s0 == "&":
            return Conjunction(pk)
        else:
            return Module(s)

    def connect(self, dst):
        self.dsts.append(dst)
        if isinstance(dst, Conjunction):
            dst.register(self)

    def send(self):
        if self.disabled:
            log.debug("%s is disabled - not sending", self.pk)
            return
        for dst in self.dsts:
            log.debug("%s -%s-> %s", self.pk, "high" if self._state else "low", dst.pk)
            pending.append((self.pk, self._state, dst.pk))
            self._n_sent[self._state] += 1

    def recv(self, pulse, sender_id):
        if self.pk == "output" or self.pk == "rx":
            self._recv.append((pulse, sender_id))
        elif self.pk == "broadcaster":
            self.send()

    def __repr__(self):
        return f"<{type(self).__name__}({self.pk}) {self._state}>"


class FlipFlop(Module):

    def recv(self, pulse, sender_id):
        if self.disabled:
            return
        if pulse:
            return
        self._state = 1 - self._state
        self.send()


class Conjunction(Module):
    def __init__(self, pk):
        super().__init__(pk)
        self.memory = {}

    def register(self, src):
        self.memory[src.pk] = 0

    def recv(self, pulse, sender_id):
        if self.disabled:
            return
        self.memory[sender_id] = pulse
        self.send()

    @property
    def _state(self):
        return 1 - all(self.memory.values())

    @_state.setter
    def _state(self, val):
        return


modules = {}
lines = sorted(data.splitlines(), key=lambda line: (line[0], len(line)))
for line in lines:
    src, _dsts = line.split(" -> ")
    mod = Module.from_str(src)
    modules[mod.pk] = mod
for line in lines:
    src, dsts = line.split(" -> ")
    src = modules[src.lstrip("%&")]
    for dst in dsts.split(", "):
        if dst not in modules:
            modules[dst] = Module(dst)
        src.connect(modules[dst])
button = modules["button"] = Module("button")
button.dsts.append(modules["broadcaster"])

for i in range(1000):
    button.send()
    while pending:
        src_pk, pulse, dst_pk = pending.popleft()
        modules[dst_pk].recv(pulse, src_pk)

n_low_sent = sum(m._n_sent[0] for m in modules.values())
n_high_sent = sum(m._n_sent[1] for m in modules.values())
a = n_low_sent * n_high_sent
print("answer_a:", a)

# dsts of broadcaster, only enable one at a time
periods = []
broadcasted = modules["broadcaster"].dsts
for branch in broadcasted:
    for m in modules.values():
        m.reset()
        if m in broadcasted and m is not branch:
            m.disabled = True
    s0 = tuple(m._state for m in modules.values())
    seen = set()
    i = 0
    while True:
        button.send()
        while pending:
            src_pk, pulse, dst_pk = pending.popleft()
            modules[dst_pk].recv(pulse, src_pk)
        i += 1
        state = tuple(m._state for m in modules.values())
        if state in seen:
            periods.append(i - 1)
            break
        seen.add(state)
b = math.lcm(*periods)
print("answer_b:", b)



"""
--- Day 2: Rock Paper Scissors ---
https://adventofcode.com/2022/day/2
"""
from enum import Enum
from functools import cached_property

from aocd import data
from bidict import bidict


class Shape(Enum):
    A = X = Rock = 1
    B = Y = Paper = 2
    C = Z = Scissors = 3

    @cached_property
    def _lookup(self):
        return bidict({
            Shape.Rock: Shape.Scissors,
            Shape.Scissors: Shape.Paper,
            Shape.Paper: Shape.Rock,
        })

    @property
    def wins_against(self):
        return self._lookup[self]

    @property
    def loses_to(self):
        return self._lookup.inv[self]


def score(glyph_elf, glyph_you, part):
    elfs_shape = Shape[glyph_elf]
    if part == "a":
        your_shape = Shape[glyph_you]
    elif part == "b":
        if glyph_you == "X":  # you should lose
            your_shape = elfs_shape.wins_against
        elif glyph_you == "Y":  # you should draw
            your_shape = elfs_shape
        elif glyph_you == "Z":  # you should win
            your_shape = elfs_shape.loses_to
    if your_shape.loses_to == elfs_shape:
        your_score = your_shape.value + 0
    elif your_shape is elfs_shape:
        your_score = your_shape.value + 3
    elif your_shape.wins_against == elfs_shape:
        your_score = your_shape.value + 6
    return your_score

a = b = 0
for line in data.splitlines():
    elf, me = line.split()
    a += score(elf, me, part="a")
    b += score(elf, me, part="b")

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 18: Like a GIF For Your Yard ---
https://adventofcode.com/2015/day/18
"""
import numpy as np
from aocd import data
from scipy.signal import convolve2d


kernel = np.ones((3, 3), dtype=int)
kernel[1, 1] = 0


def evolve(A, part="a"):
    """conway's game of life"""
    C = convolve2d(A, kernel, mode="same")
    A = ((A == 1) & ((C == 2) | (C == 3))) | ((A == 0) & (C == 3))
    return A.astype(int)


def animate(data, part="a"):
    A = np.array([[c == "#" for c in r] for r in data.splitlines()]).astype(int)
    if part == "b":
        A[0, 0] = A[0, -1] = A[-1, 0] = A[-1, -1] = 1
    iterations = 1 if A.shape == (6, 6) else 100
    for i in range(iterations):
        A = evolve(A, part)
        if part == "b":
            A[0, 0] = A[0, -1] = A[-1, 0] = A[-1, -1] = 1
    return A.sum()


print("answer_a:", animate(data, part="a"))
print("answer_b:", animate(data, part="b"))



"""
--- Day 13: Packet Scanners ---
https://adventofcode.com/2017/day/13
"""
from aocd import data


depth2range = dict(map(int, line.split(":")) for line in data.splitlines())
print("answer_a:", sum(d * r for d, r in depth2range.items() if d % (2 * r - 2) == 0))

t = 0
while not all((d + t) % (2 * r - 2) for d, r in depth2range.items()):
    t += 1
print("answer_b:", t)



"""
--- Day 8: I Heard You Like Registers ---
https://adventofcode.com/2017/day/8
"""
import operator as op
from collections import defaultdict
from types import SimpleNamespace

from aocd import data
from parse import parse


ops = {
    ">": op.gt,
    "<": op.lt,
    ">=": op.ge,
    "<=": op.le,
    "==": op.eq,
    "!=": op.ne,
    "inc": op.iadd,
    "dec": op.isub,
}


b = 0
d = defaultdict(int)
template = "{x} {i:op} {m:d} if {y} {cmp:op} {n:d}"
for line in data.splitlines():
    parsed = parse(template, line, {"op": ops.get})
    r = SimpleNamespace(**parsed.named)
    d[r.x] = r.i(d[r.x], r.cmp(d[r.y], r.n) and r.m)
    b = max(b, d[r.x])


print("answer_a:", max(d.values()))
print("answer_b:", b)



"""
--- Day 6: Probably a Fire Hazard ---
https://adventofcode.com/2015/day/6
"""
import numpy as np
from aocd import data


A = np.zeros((1000, 1000), dtype=bool)
B = np.zeros((1000, 1000), dtype=int)

for line in data.splitlines():
    action, p1, _, p2 = line.rsplit(None, 3)
    x1, y1 = [int(n) for n in p1.split(",")]
    x2, y2 = [int(n) for n in p2.split(",")]
    t = slice(x1, x2 + 1), slice(y1, y2 + 1)
    if action == "toggle":
        A[t] = ~A[t]
        B[t] += 2
    elif action == "turn on":
        A[t] = True
        B[t] += 1
    elif action == "turn off":
        A[t] = False
        B[t] -= 1
        B = B.clip(min=0)

print("answer_a:", A.sum())
print("answer_b:", B.sum())



"""
--- Day 23: Category Six ---
https://adventofcode.com/2019/day/23
"""
import logging

from aocd import data

from aoc_wim.aoc2019 import IntComputer


log = logging.getLogger(__name__)


class Network:
    def __init__(self, data):
        self.y0 = set()
        self.xy = None, None
        self.comps = [IntComputer(data, inputs=[i]) for i in range(50)]
        self.done = False

    def step(self):
        for i, comp in enumerate(self.comps):
            try:
                comp.run(until=IntComputer.op_output)
            except IndexError:
                # attempted pop from an empty deque
                comp.input.appendleft(-1)
                comp.step()
            if len(comp.output) >= 3:
                # 3 outputs: packet ready to consume/transmit
                dest = comp.output.pop()
                x = comp.output.pop()
                y = comp.output.pop()
                log.info("comp[%d] sends (x=%d, y=%d) to comp[%d]", i, x, y, dest)
                if dest == 0xFF:
                    self.xy = x, y
                    if any(c.output or c.input for c in self.comps):
                        continue
                    if not self.y0:
                        print("answer_a:", y)
                    if y in self.y0:
                        print("answer_b:", y)
                        nic.done = True
                    dest = 0
                    self.y0.add(y)
                self.comps[dest].input.appendleft(x)
                self.comps[dest].input.appendleft(y)

    def run(self):
        while not self.done:
            nic.step()


nic = Network(data)
nic.run()



"""
--- Day 7: Handy Haversacks ---
https://adventofcode.com/2020/day/7
"""
from collections import deque

import networkx as nx
from aocd import data

g = nx.DiGraph()
for line in data.splitlines():
    left, right = line.split("s contain ")
    rights = right.split(", ")
    for right in rights:
        n, right = right.split(None, 1)
        right = right.rstrip("s.")
        n = 0 if n == "no" else int(n)
        g.add_edge(left, right, weight=n)

print("answer_a:", sum(nx.has_path(g, b, "shiny gold bag") for b in g.nodes) - 1)

b = -1
q = deque([(1, "shiny gold bag")])
while q:
    w, b0 = q.popleft()
    b += w
    q.extend((w * g[b0][b1]["weight"], b1) for b1 in g[b0])

print("answer_b:", b)



"""
--- Day 16: Proboscidea Volcanium ---
https://adventofcode.com/2022/day/16
"""
import itertools
from collections import defaultdict
from collections import deque

from aocd import data
from parse import parse


flow = {}
template = "Valve {} ha flow rate={:d}; tunnel lead to valve {}"
dist = defaultdict(lambda: float("inf"))
for line in data.splitlines():
    line = line.replace("s ", " ")
    v1, f, v2s = parse(template, line).fixed
    flow[v1] = f
    dist[v1, v1] = 0
    for v2 in v2s.split(", "):
        dist[v1, v2] = dist[v2, v1] = 1

# https://en.wikipedia.org/wiki/Floyd-Warshall_algorithm
for k, i, j in itertools.product(flow, repeat=3):
    dist[i, j] = min(dist[i, j], dist[i, k] + dist[k, j])

nonzero_flow = {k: v for k, v in flow.items() if v > 0}


def find_paths():
    # state vector:
    # current minute (int), current score (int), current valve (str), {valves_opened: minute} (dict)
    s0 = 0, 0, "AA", {}
    result = {}
    q = deque([s0])
    while q:
        t, s, pos, visited = q.popleft()
        k = frozenset(visited)
        if k not in result or result[k][1] < s:
            result[k] = visited, s
        for p in nonzero_flow:
            if p in visited:
                continue
            d = dist[pos, p]
            dt = 30 - t - d - 1
            if dt > 0:
                q.append((t + d + 1, s + flow[p] * dt, p, {**visited, p: t + d + 1}))
    return result.values()


paths = find_paths()
print("answer_a:", max(s for v, s in paths))

scores = {}
for path, _old_score in paths:
    # just prune the paths we already had for max_t=30
    path = {k: v for k, v in path.items() if v < 26}
    k = frozenset(path)
    # scores need to be recalculated
    score = sum((26 - t) * flow[k] for k, t in path.items())
    if scores.get(k, -1) < score:
        scores[k] = score

# order the paths with highest pressure relief first
visited = sorted(scores, key=scores.get, reverse=True)
b = 0
for i in range(len(visited)):
    v1 = visited[i]
    for j in range(i + 1, len(visited)):
        v2 = visited[j]
        if not v1 & v2:
            b = max(b, scores[v1] + scores[v2])
            break
    if scores[v1] * 2 <= b:
        # short-circuit because we can not possibly do any better now
        break

print("answer_b:", b)



"""
--- Day 2: Dive! ---
https://adventofcode.com/2021/day/2
"""
from aocd import data

from aoc_wim.zgrid import ZGrid

aim = za = zb = 0
for line in data.splitlines():
    direction, X = line.split()
    X = int(X)
    dz = ZGrid.dzs[direction] * X
    za += dz
    if direction == "forward":
        zb += dz + aim * X
    else:
        aim += dz

print("answer_a:", int(za.real) * int(za.imag))
print("answer_b:", int(zb.real) * int(zb.imag))



"""
--- Day 2: Cube Conundrum ---
https://adventofcode.com/2023/day/2
"""
from collections import Counter
from math import prod

from aocd import data

bag = Counter("rgb"*12 + "g" + "bb")
a = b = 0
for line in data.replace(";", ",").splitlines():
    g, line = line.removeprefix("Game").split(":")
    d = Counter()
    for n_c in line.split(","):
        n, c = n_c.split()
        d[c[0]] = max(d[c[0]], int(n))
    a += int(g) * (d <= bag)
    b += prod(d.values())

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 1: Not Quite Lisp ---
https://adventofcode.com/2015/day/1
"""
from aocd import data


direction = {"(": +1, ")": -1}
basement = None
floor = 0
for i, c in enumerate(data, 1):
    floor += direction[c]
    if basement is None and floor == -1:
        basement = i


print("answer_a:", floor)
print("answer_b:", basement)



"""
--- Day 20: Grove Positioning System ---
https://adventofcode.com/2022/day/20
"""
from aocd import data


for key, rounds, part in zip((1, 811589153), (1, 10), "ab"):
    ns = [int(n) * key for n in data.split()]
    N = len(ns)
    idx = [*range(N)]
    for _ in range(rounds):
        for i in range(N):
            idx.pop(pos := idx.index(i))
            idx.insert((pos + ns[i]) % (N - 1), i)
    i0 = idx.index(ns.index(0))  # where did 0 get 'mixed' to?
    result = sum(ns[idx[(i0 + i) % N]] for i in [1000, 2000, 3000])
    print(f"answer_{part}:", result)



"""
--- Day 1: Sonar Sweep ---
https://adventofcode.com/2021/day/1
"""
from aocd import data

ns = [int(n) for n in data.split()]
print("answer_a:", sum(n2 > n1 for n1, n2 in zip(ns, ns[1:])))
print("answer_b:", sum(n2 > n1 for n1, n2 in zip(ns, ns[3:])))



"""
--- Day 20: Infinite Elves and Infinite Houses ---
https://adventofcode.com/2015/day/20
"""
import numpy as np
from aocd import data


A = np.zeros(1000000, dtype=int)
B = A.copy()
for i in range(1, 1000000):
    A[i::i] += i * 10
    B[i::i][:50] += i * 11

print("answer_a:", np.argmax(A > int(data or 0)))
print("answer_b:", np.argmax(B > int(data or 0)))



"""
--- Day 7: The Sum of Its Parts ---
https://adventofcode.com/2018/day/7
"""
import logging
from types import SimpleNamespace

from aocd import data
from parse import parse


log = logging.getLogger(__name__)
# logging.basicConfig(format='%(message)s', level=logging.DEBUG)


class Worker:
    def __init__(self):
        self.has = "."  # idle
        self.t = -1

    def tick(self):
        self.t -= 1
        if self.t == 0:
            val = self.has
            self.has = "."
            return val


def work(data, part="a"):
    template = "Step {} must be finished before step {} can begin."
    pairs = [parse(template, s).fixed for s in data.splitlines()]
    remaining = {x for pair in pairs for x in pair}
    n_workers = 1 if part == "a" else 5
    delay = 60
    if remaining == set("ABCDEF"):
        # the smaller example data
        n_workers = 1 if part == "a" else 2
        delay = 0
    in_progress = set()
    done = set()
    text = ""
    t = -1
    workers = [Worker() for i in range(n_workers)]
    while remaining or in_progress:
        for w in workers:
            x = w.tick()
            if x is not None:
                text += x
                pairs = [pair for pair in pairs if x not in pair]
                in_progress.remove(x)
                done.add(x)
        avail = remaining - {two for one, two in pairs} - in_progress
        log.debug("candidates %d: %s", t, avail)
        for w in workers:
            if w.has == "." and avail:
                w.has = x = min(avail)
                avail.remove(x)
                remaining.remove(x)
                in_progress.add(x)
                if n_workers == 1:
                    w.t = 1
                else:
                    w.t = ord(x) - ord("A") + 1 + delay
        t += 1
        log.debug("%4d: %s   %s", t, " ".join(w.has for w in workers), text)
    result = SimpleNamespace(text=text, n_iterations=t)
    return result


print("answer_a:", work(data).text)
print("answer_b:", work(data, part="b").n_iterations)



"""
--- Day 7: Amplification Circuit ---
https://adventofcode.com/2019/day/7
"""
from itertools import permutations

from aocd import data

from aoc_wim.aoc2019 import IntComputer


class Amp:
    def __init__(self, data, phase_settings):
        self.comps = [IntComputer(data, inputs=[val]) for val in phase_settings]
        for left, right in zip(self.comps, self.comps[1:] + [self.comps[0]]):
            left.output = right.input

    def run(self):
        self.comps[0].input.appendleft(0)
        while True:
            for comp in self.comps:
                try:
                    comp.run(until=IntComputer.op_output)
                except IntComputer.Halt:
                    return self.comps[-1].output[-1]


print("answer_a:", max(Amp(data, p).run() for p in permutations(range(5))))
print("answer_b:", max(Amp(data, p).run() for p in permutations(range(5, 10))))



"""
--- Day 19: Tractor Beam ---
https://adventofcode.com/2019/day/19
"""
import functools

from aocd import data

from aoc_wim.aoc2019 import IntComputer
from aoc_wim.search import Bisect
from aoc_wim.zgrid import ZGrid


@functools.lru_cache(maxsize=100**2)
def beam(z):
    comp = IntComputer(data, inputs=[int(z.imag), int(z.real)])
    comp.run(until=IntComputer.op_output)
    [result] = comp.output
    return result


def left_edge_of_beam(y, gradient, beam=beam):
    x = int(y / gradient)
    z = x + y*1j
    if beam(z):
        while beam(z - 1):
            z -= 1
    else:
        while not beam(z + 1):
            z += 1
        z += 1
    assert beam(z) and not beam(z - 1)
    return z


def locate_square(beam, width, gradient_estimate=1., hi=None):
    d = width - 1

    def check(y):
        z = left_edge_of_beam(y, gradient_estimate, beam)
        val = beam(z + d * ZGrid.NE)
        print(f"y={y}", "wide" if val else "narrow")
        return val

    bisect = Bisect(check, lo=d, hi=hi)
    print("bisecting...")
    y = bisect.run() + 1
    z = left_edge_of_beam(y, gradient_estimate, beam) + d * ZGrid.N
    return z


if __name__ == "__main__":
    print("populating 50x50 zgrid...")
    grid = ZGrid()
    x0 = 0
    for y in range(50):
        on = False
        for x in range(x0, 50):
            z = x + y * 1j
            val = grid[z] = beam(z)
            if not on and val:
                on = True
                x0 = x
                if x0:
                    m = y / x0
            if on and not val:
                break

    print("answer_a:", sum(grid.values()))
    grid.translate({0: ".", 1: "#"})
    grid.draw()

    print("initial gradient is approx -->", m)
    print("refining gradient estimate -->", end=" ")
    z = left_edge_of_beam(2000, gradient=m)
    m = z.imag/z.real
    print(m)

    z = locate_square(beam, width=100, gradient_estimate=m)
    print("answer_b:", int(z.real)*10000 + int(z.imag))



"""
--- Day 16: The Floor Will Be Lava ---
https://adventofcode.com/2023/day/16
"""
from aocd import data

from aoc_wim.zgrid import ZGrid


def part_a(z0, dz0):
    paths = []
    rays = [[(z0, dz0)]]
    seen = set()
    while rays:
        ray = rays[0]
        if len(ray) > 1 and ray[0] == ray[-1]:
            paths.append([z for z, dz in ray])
            del rays[0]
            continue
        z, dz = ray[-1]
        z += dz
        if (z, dz) in seen:
            paths.append([z for z, dz in ray])
            del rays[0]
            continue
        seen.add((z, dz))
        if z not in grid:
            paths.append([z for z, dz in ray])
            del rays[0]
            continue
        match grid[z], dz.real, dz.imag:
            case  ".", _, _: ray.append((z, dz))
            case  "-", _, 0: ray.append((z, dz))
            case  "|", 0, _: ray.append((z, dz))
            case  "/", _, 0: ray.append((z, grid.turn_left * dz))
            case  "/", 0, _: ray.append((z, grid.turn_right * dz))
            case "\\", _, 0: ray.append((z, grid.turn_right * dz))
            case "\\", 0, _: ray.append((z, grid.turn_left * dz))
            case "|", _, 0:
                paths.append([z for z, dz in ray])
                del rays[0]
                if z + grid.up in grid:
                    rays.append([(z, grid.up)])
                if z + grid.down in grid:
                    rays.append([(z, grid.down)])
            case "-", 0, _:
                paths.append([z for z, dz in ray])
                del rays[0]
                if z + grid.left in grid:
                    rays.append([(z, grid.left)])
                if z + grid.right in grid:
                    rays.append([(z, grid.right)])
    return len({z for z, dz in seen if z in grid})


grid = ZGrid(data)
print("answer_a:", part_a(z0=-1, dz0=1))

w, h = grid.width, grid.height
bs = []
for col in range(w):
    bs.append(part_a(z0=complex(col, -1j), dz0=grid.down))
    bs.append(part_a(z0=complex(col, h), dz0=grid.up))
for row in range(h):
    bs.append(part_a(z0=complex(-1, row), dz0=grid.right))
    bs.append(part_a(z0=complex(w, row), dz0=grid.left))
print("answer_b:", max(bs))



"""
--- Day 13: Distress Signal ---
https://adventofcode.com/2022/day/13
"""
from functools import cmp_to_key
from json import loads

from aocd import data


def cmp(a, b):
    match a, b:
        case int(), int():
            return a - b
        case list(), list():
            for x, y in zip(a, b):
                if c := cmp(x, y):
                    return c
            return cmp(len(a), len(b))
        case list(), int():
            return cmp(a, [b])
        case int(), list():
            return cmp([a], b)


a = 0
packets = []
for i, chunk in enumerate(data.split("\n\n"), 1):
    first, second = map(loads, chunk.splitlines())
    packets += [first, second]
    if cmp(first, second) < 0:
        a += i

packets += [[], [[2]], [[6]]]
packets.sort(key=cmp_to_key(cmp))
b = packets.index([[2]]) * packets.index([[6]])

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 24: Electromagnetic Moat ---
https://adventofcode.com/2017/day/24
"""
from aocd import data


def extend(bridge, segments):
    *rest, (connected, free) = bridge
    end = True
    for i, segment in enumerate(segments):
        if free in segment:
            end = False
            next_segments = segments[:i] + segments[i + 1 :]
            next_bridge = bridge + [(free, segment[segment[0] == free])]
            yield from extend(next_bridge, next_segments)
    if end:
        yield bridge


segments = [tuple(int(x) for x in line.split("/")) for line in data.splitlines()]
bridges = list(extend([(0, 0)], segments))
lengths = [len(b) for b in bridges]
strengths = [sum(x + y for x, y in b) for b in bridges]
longest = max(lengths)
long_indices = [i for i, length in enumerate(lengths) if length == longest]

print("answer_a:", max(strengths))
print("answer_b:", max([strengths[i] for i in long_indices]))



"""
--- Day 22: Reactor Reboot ---
https://adventofcode.com/2021/day/22
"""
from dataclasses import dataclass

from aocd import data
from parse import parse


@dataclass
class Cuboid:
    # 3D space of {x,y,z} with
    #   x0 <= x <= x1
    #   y0 <= y <= y1
    #   z0 <= z <= z1

    x0: int
    x1: int
    y0: int
    y1: int
    z0: int
    z1: int
    val: int

    @classmethod
    def from_line(cls, line):
        template = "{} x={:d}..{:d},y={:d}..{:d},z={:d}..{:d}"
        v, x0, x1, y0, y1, z0, z1 = parse(template, line).fixed
        return cls(x0, x1, y0, y1, z0, z1, 1 if v == "on" else -1)

    def __and__(self, other):
        if not isinstance(other, Cuboid):
            return NotImplemented
        x0 = self.x0 if self.x0 > other.x0 else other.x0
        y0 = self.y0 if self.y0 > other.y0 else other.y0
        z0 = self.z0 if self.z0 > other.z0 else other.z0
        x1 = self.x1 if self.x1 < other.x1 else other.x1
        y1 = self.y1 if self.y1 < other.y1 else other.y1
        z1 = self.z1 if self.z1 < other.z1 else other.z1
        if x0 <= x1 and y0 <= y1 and z0 <= z1:
            return Cuboid(x0, x1, y0, y1, z0, z1, -self.val)


def total_volume(cuboids, part="a"):
    if part == "a":
        cuboids = [c for c in cuboids if all(abs(v) <= 50 for v in vars(c).values())]
    cs = []
    for cuboid in cuboids:
        # https://en.wikipedia.org/wiki/Inclusion-exclusion_principle
        extra = [c & cuboid for c in cs]
        cs += [c for c in extra if c is not None]
        if cuboid.val == 1:
            cs.append(cuboid)
    return sum((c.x1 - c.x0 + 1) * (c.y1 - c.y0 + 1) * (c.z1 - c.z0 + 1) * c.val for c in cs)


cuboids = [Cuboid.from_line(x) for x in data.splitlines()]
print("answer_a:", total_volume(cuboids, part="a"))
print("answer_b:", total_volume(cuboids, part="b"))



"""
--- Day 9: Marble Mania ---
https://adventofcode.com/2018/day/9
"""
from aocd import data
from parse import parse


class Node:

    __slots__ = "val", "left", "right"

    def __init__(self, val, left=None, right=None):
        self.val = val
        self.left = left
        self.right = right


def get_high_score(n_players, n_marbles):
    n = Node(val=0)
    n.left = n.right = n
    scores = [0] * n_players
    for m in range(n_marbles):
        if m % 23 != 22:
            left = n.right
            right = n.right.right
            n = Node(val=m + 1, left=left, right=right)
            left.right = right.left = n
        else:
            n = right = n.left.left.left.left.left.left
            left = n.left.left
            scores[m % n_players] += m + 1 + n.left.val
            left.right = right
            right.left = left
    high_score = max(scores)
    return high_score


template = "{:d} players; last marble is worth {:d} points"
n_players, n_marbles = parse(template, data)
print("answer_a:", get_high_score(n_players, n_marbles))
print("answer_b:", get_high_score(n_players, n_marbles * 100))



"""
--- Day 14: Chocolate Charts ---
https://adventofcode.com/2018/day/14
"""
from aocd import data


class RecipeList:

    def __init__(self, initial=(3, 7)):
        self.scores = list(initial)
        self.scores_str = "".join([str(x) for x in self.scores])
        self.elf1 = 0
        self.elf2 = 1

    def grow(self, min_length):
        scores = self.scores
        e1, e2 = self.elf1, self.elf2
        prev_len = len(scores)
        while len(scores) < min_length:
            score1 = scores[e1]
            score2 = scores[e2]
            score = score1 + score2
            if score >= 10:
                scores.append(1)
                scores.append(score - 10)
            else:
                scores.append(score)
            e1 = (e1 + 1 + score1) % len(scores)
            e2 = (e2 + 1 + score2) % len(scores)
        self.elf1 = e1
        self.elf2 = e2
        self.scores_str += "".join([str(x) for x in scores[prev_len:]])


recipes = RecipeList()

k = int(data)
if k < 10**6:
    recipes.grow(k + 10)
    print("answer_a:", "".join(str(n) for n in recipes.scores[k:k + 10]))

chunk = 2_000_000
i = -1
while i == -1:
    recipes.grow(min_length=len(recipes.scores) + chunk)
    i = recipes.scores_str.find(data, -chunk - len(data))
print("answer_b:", i)



"""
--- Day 9: Rope Bridge ---
https://adventofcode.com/2022/day/9
"""
from aocd import data


def cmp(x, y):
    return (x > y) - (x < y)


dHs = dict(zip("UDLR", (-1j, 1j, -1, 1)))
seen_a = set()
seen_b = set()
zs = [0] * 10
for line in data.splitlines():
    dH, n = line.split()
    for _ in range(int(n)):
        zs[0] += dHs[dH]
        for i in range(9):
            dz = zs[i] - zs[i + 1]
            if abs(dz) >= 2:
                zs[i + 1] += cmp(dz.real, 0) + 1j * cmp(dz.imag, 0)
        seen_a.add(zs[1])
        seen_b.add(zs[-1])

print("answer_a:", len(seen_a))
print("answer_b:", len(seen_b))



"""
--- Day 24: Air Duct Spelunking ---
https://adventofcode.com/2016/day/24
"""
from itertools import combinations
from itertools import permutations

import networkx as nx
from aocd import data

from aoc_wim.zgrid import ZGrid

grid = ZGrid(data, on=".", off="#")
graph = grid.graph(extra="0123456789")
distances = {}
for a, b in combinations(graph.extra, 2):
    d = nx.shortest_path_length(graph, graph.extra[a], graph.extra[b])
    distances[a, b] = distances[b, a] = d


def shortest_path(part="a"):
    paths = {}
    for nodes in permutations(graph.extra.keys() - {"0"}):
        path = "0" + "".join(nodes)
        if part == "b":
            path += "0"
        paths[path] = sum(distances[a, b] for a, b in zip(path, path[1:]))
    return min(paths.values())


print("answer_a:", shortest_path(part="a"))
print("answer_b:", shortest_path(part="b"))



"""
--- Day 17: Conway Cubes ---
https://adventofcode.com/2020/day/17
"""
from itertools import product

import numpy as np
from aocd import data


def evolve(A):
    A0 = np.pad(A, pad_width=1)
    A1 = A0.copy()
    for pos in [*product(*[range(d) for d in A0.shape])]:
        slices = [slice(max(x-1, 0), x+2) for x in pos]
        n_on = A0[tuple(slices)].sum()
        if A0[pos] and not 3 <= n_on <= 4:
            A1[pos] = 0
        if not A0[pos] and n_on == 3:
            A1[pos] = 1
    return A1


A0 = np.array([[v == "#" for v in line] for line in data.splitlines()], dtype=int)
for dimension, part in enumerate("ab", start=3):
    A = A0.copy()[(...,) + (None,) * (dimension - A0.ndim)]
    for _ in range(6):
        A = evolve(A)
    print(f"answer_{part}:", A.sum())



"""
--- Day 2: Inventory Management System ---
https://adventofcode.com/2018/day/2
"""
from collections import Counter
from itertools import combinations

from aocd import data


counters = [Counter(s) for s in data.split()]
doubles = sum(1 for c in counters if 2 in c.values())
triples = sum(1 for c in counters if 3 in c.values())
print("answer_a:", doubles * triples)

for a, b in combinations(data.split(), 2):
    s = "".join([x for x, y in zip(a, b) if x == y])
    if len(s) == len(a) - 1:
        print("answer_b:", s)
        break



"""
--- Day 9: Mirage Maintenance ---
https://adventofcode.com/2023/day/9
"""
import numpy as np
from aocd import data


n = len(data.splitlines())
A = A0 = np.fromstring(data, sep=" ", dtype=int).reshape(n, -1).T
As = [A0]
while A.any():
    A = A[1:] - A[:-1]
    As.append(A)

As.reverse()
Ba = Bb = np.pad(A, [(0, 1), (0, 0)])
for A0, A1 in zip(As, As[1:]):
    Ba = np.vstack([A0[:1], A1 + Ba])
    Bb = np.vstack([A1 - Bb, A0[-1:]])

print("answer_a:", int(Ba[-1].sum()))
print("answer_b:", int(Bb[0].sum()))



"""
--- Day 4: Scratchcards ---
https://adventofcode.com/2023/day/4
"""
from aocd import data

lines = data.splitlines()
cards = [1] * len(lines)
a = 0
for i, line in enumerate(lines):
    L, R = line.split("|")
    w = {*L.split()} & {*R.split()}
    a += int(2 ** (len(w) - 1))
    for j in range(min(len(w), len(cards))):
        cards[i + 1 + j] += cards[i]

print("answer_a:", a)
print("answer_b:", sum(cards))



"""
--- Day 21: Monkey Math ---
https://adventofcode.com/2022/day/21
"""
import operator as op

from aocd import data

ops = {
    "+": op.add,
    "*": op.mul,
    "-": op.sub,
    "/": op.truediv,
}


def solve(eqns):
    known = {}
    unknown = eqns.copy()
    while unknown:
        for name, eqn in unknown.items():
            if eqn in known or isinstance(eqn, complex) or eqn.isdigit():
                known[name] = complex(known.get(eqn, eqn))
                del unknown[name]
                break
            e1, o, e2 = eqn.split()
            if e1 in known or e1.isdigit():
                e1 = complex(known.get(e1, e1))
            if e2 in known or e2.isdigit():
                e2 = complex(known.get(e2, e2))
            if type(e1) is type(e2) is complex:
                known[name] = ops[o](e1, e2)
                del unknown[name]
                break
    return known


eqns = dict(line.split(": ") for line in data.splitlines())
known = solve(eqns)
print("answer_a:", int(known["root"].real))

eqns["humn"] = 1j
m1, _, m2 = eqns.pop("root").split()
known = solve(eqns)
c1 = known[m1]
c2 = known[m2]
print("answer_b:", int((c1.real - c2.real) / (c2.imag - c1.imag)))



"""
--- Day 24: Lobby Layout ---
https://adventofcode.com/2020/day/24
"""
import re

import numpy as np
from aocd import data
from scipy.signal import convolve2d

from aoc_wim import zgrid


pat = r"(se|sw|ne|nw|e|w)"
grid = zgrid.ZGrid(on=1, off=0)
for i, line in enumerate(data.splitlines()):
    steps = re.findall(pat, line)
    z = sum([zgrid.hexH[s] for s in steps])
    grid[z] = 1 - grid.get(z, 0)
    # grid.draw_hex(glyph=0, orientation="H", clear=True, title=f" flip {i} ")
print("answer_a:", grid.count(1))

# conway's game of life on a hexgrid
A = np.array(grid)
kernel = 1 - np.flipud(np.eye(3, dtype=int))
for day in range(1, 101):
    A = np.pad(A, pad_width=1)
    C = convolve2d(A, kernel, mode="same")
    A = ((A == 1) & ((C == 1) | (C == 2))) | ((A == 0) & (C == 2))
    A = A.astype(int)
    if day <= 10 or day % 10 == 0:
        print(f"Day {day}:", A.sum())
    if day == 10:
        print()
print("answer_b:", A.sum())



"""
--- Day 10: Adapter Array ---
https://adventofcode.com/2020/day/10
"""
from aocd import data

ns = [*map(int, data.split())]
ns.sort()
d1 = d3 = prev = 0
for n in ns:
    d1 += n - prev == 1
    d3 += n - prev == 3
    prev = n
d3 += 1  # extra hop to device
print("answer_a:", d1 * d3)

memo = {0: 1}
for n in ns:
    memo[n] = sum(memo.get(n - i, 0) for i in range(4))

print("answer_b:", memo[n])



"""
--- Day 17: Reservoir Research ---
https://adventofcode.com/2018/day/17
"""
from aocd import data


def parsed(data):
    grid = {}
    for line in data.splitlines():
        a, b = line.split(", ")
        a, n = a.split("=")
        if a == "x":
            xs = [int(n)]
            y, start_stop = b.split("=")
            start, stop = start_stop.split("..")
            ys = range(int(start), int(stop) + 1)
        else:
            ys = [int(n)]
            x, start_stop = b.split("=")
            start, stop = start_stop.split("..")
            xs = range(int(start), int(stop) + 1)
        for y in ys:
            for x in xs:
                grid[complex(x, y)] = "#"
    ys = [z.imag for z in grid]
    ymin, ymax = int(min(ys)), int(max(ys))
    grid[0j + 500] = "+"
    return grid, ymin, ymax


def draw(grid, pause=False):
    print("\33c")
    xs = [int(z.real) for z in grid]
    xrange = range(min(xs) - 1, max(xs) + 2)
    yrange = range(0, int(max(z.imag for z in grid)) + 1)
    for y in yrange:
        line = []
        for x in xrange:
            line.append(grid.get(complex(x, y), "."))
        line = "".join(line)
        print(line)
        if not set(line).intersection("|+~"):
            break
    print()
    if pause:
        input("Press enter to continue...")


def xflow(grid, pos0):
    # find bounds left and right
    pos = pos0
    boundL = boundR = None
    next_flow = set()
    while True:
        if grid.get(pos - 1) == "#":
            boundL = pos
            break
        if grid.get(pos - 1 + 1j) == "|":
            break
        if grid.get(pos - 1, "|") == "|" and pos - 1 + 1j not in grid:
            next_flow.add(pos - 1)
            break
        pos -= 1
        grid[pos] = "|"
    pos = pos0
    while True:
        if grid.get(pos + 1) == "#":
            boundR = pos
            break
        if grid.get(pos + 1 + 1j) == "|":
            break
        if grid.get(pos + 1, "|") == "|" and pos + 1 + 1j not in grid:
            next_flow.add(pos + 1)
            break
        pos += 1
        grid[pos] = "|"
    if boundL is not None and boundR is not None:
        pos = boundL
        while True:
            grid[pos] = "~"
            pos += 1
            if pos.real > boundR.real:
                break
        next_flow.add(pos0 - 1j)
    return next_flow


def wet(grid, ymax, tap=1j + 500):
    flow = {tap}
    while flow:
        next_flow = set()
        for c in flow:
            grid[c] = "|"
            if c.imag >= ymax:
                continue
            if c + 1j not in grid:
                next_flow |= {c + 1j}
            elif grid[c + 1j] == "|":
                continue
            else:
                next_flow |= xflow(grid, c)
        flow = next_flow


grid, ymin, ymax = parsed(data)
wet(grid, ymax)
s = "".join(grid.values())
b = s.count("~")
a = s.count("|") + b - ymin + 1
print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 22: Sand Slabs ---
https://adventofcode.com/2023/day/22
"""
from aocd import data
from dataclasses import dataclass
import operator as op


@dataclass
class Brick:
    x0: int
    y0: int
    z0: int
    x1: int
    y1: int
    z1: int
    id: int = None

    @classmethod
    def fromline(cls, line):
        return cls(*map(int, line.replace("~", ",").split(",")))

    def __post_init__(self):
        if self.x0 > self.x1 or self.y0 > self.y1 or self.z0 > self.z1:
            raise ValueError(f"Invalid brick {self}")

    def cells(self, dz=0):
        xs = range(self.x0, self.x1 + 1)
        ys = range(self.y0, self.y1 + 1)
        zs = range(self.z0, self.z1 + 1)
        return {(x, y, z + dz) for x in xs for y in ys for z in zs}

    def above(self):
        xs = range(self.x0, self.x1 + 1)
        ys = range(self.y0, self.y1 + 1)
        return {(x, y, self.z1 + 1) for x in xs for y in ys}

    def below(self):
        xs = range(self.x0, self.x1 + 1)
        ys = range(self.y0, self.y1 + 1)
        return {(x, y, self.z0 - 1) for x in xs for y in ys}

    def fall(self):
        self.z0 -= 1
        self.z1 -= 1


tdata = """1,0,1~1,2,1
0,0,2~2,0,2
0,2,3~2,2,3
0,0,4~0,2,4
2,0,5~2,2,5
0,1,6~2,1,6
1,1,8~1,1,9"""
# data = tdata

bricks = [Brick.fromline(x) for x in data.splitlines()]
bricks.sort(key=op.attrgetter("z0"))
space = {}
for i, brick in enumerate(bricks):
    brick.id = i
    xs = range(brick.x0, brick.x1 + 1)
    ys = range(brick.y0, brick.y1 + 1)
    zs = range(brick.z0, brick.z1 + 1)
    space.update(dict.fromkeys(brick.cells(), brick.id))

while True:
    mutated = False
    for brick in bricks:
        if brick.z0 == 1:
            continue
        if not space.keys() & brick.below():
            for cell in brick.cells():
                del space[cell]
            brick.fall()
            space.update()
            for cell in brick.cells():
                space[cell] = brick.id
            mutated = True
    if not mutated:
        break

bricks_below = [set() for b in bricks]
bricks_above = [set() for b in bricks]
for brick in bricks:
    for cell in brick.above():
        if cell in space:
            bricks_above[brick.id].add(space[cell])
            bricks_below[space[cell]].add(brick.id)

supports = [set() for b in bricks]
for brick in bricks:
    for brick_above in bricks_above[brick.id]:
        if len(bricks_below[brick_above]) == 1:
            supports[brick.id].add(brick_above)

a = sum(not s for s in supports)
print("answer_a:", a)

b = 0
for brick in bricks:
    stack = list(supports[brick.id])
    cumulative_support = set()
    while stack:
        b_id = stack.pop()
        cumulative_support.add(b_id)
        for brick_above in bricks_above[b_id]:
            if bricks_below[brick_above] <= cumulative_support:
                stack.append(brick_above)
    b += len(cumulative_support)
print("answer_b:", b)



"""
--- Day 1: No Time for a Taxicab ---
https://adventofcode.com/2016/day/1
"""
from aocd import data

from aoc_wim.zgrid import manhattan_distance


z = 0
dz = -1j
seen = {z}
turns = {"R": 1j, "L": -1j}
b = None
for step in data.split(", "):
    turn, n_blocks = step[0], int(step[1:])
    dz *= turns[turn]
    for block in range(n_blocks):
        z += dz
        if b is None and z in seen:
            b = manhattan_distance(z)
        seen.add(z)

print("answer_a:", manhattan_distance(z))
print("answer_b:", b)



"""
--- Day 13: A Maze of Twisty Little Cubicles ---
https://adventofcode.com/2016/day/13
"""
from functools import partial

from aocd import data

from aoc_wim.zgrid import ZGrid


def wall(z, fav_number):
    if z.real < 0 or z.imag < 0:
        return "#"
    x, y = z.real, z.imag
    fz = x * x + 3 * x + 2 * x * y + y + y * y
    popcount = bin(int(fz) + fav_number).count("1")
    result = "#" if popcount % 2 else "."
    return result


def make_grid(data):
    func = partial(wall, fav_number=int(data))
    grid = ZGrid(func, on=".", off="#")
    return grid


z0 = 1 + 1j
target = 7 + 4j if data == "10" else 31 + 39j


if __name__ == "__main__":
    grid = make_grid(data)
    depths = grid.bfs(target=target, z0=z0)
    print("answer_a:", depths[target])
    depths = grid.bfs(z0=z0, max_depth=50)
    print("answer_b:", len(depths))
    grid.draw_path(z=target, z0=z0)



"""
--- Day 3: Rucksack Reorganization ---
https://adventofcode.com/2022/day/3
"""
from itertools import batched

from aocd import data

a = 0
for line in data.splitlines():
    i = len(line) // 2  # midpoint
    left, right = line[:i], line[i:]
    [c] = set(left).intersection(right)
    a += c.isupper() * 26
    a += ord(c.lower()) - ord("a") + 1

b = 0
for s1, s2, s3 in batched(data.splitlines(), 3):
    [c] = set(s1).intersection(s2, s3)
    b += c.isupper() * 26
    b += ord(c.lower()) - ord("a") + 1

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 22: Wizard Simulator 20XX ---
https://adventofcode.com/2015/day/22
"""
from aocd import data
from fields import Fields

from aoc_wim.search import AStar


class Spell(Fields.name.cost.duration[0].damage[0].healing[0].armor[0].mana[0]):
    pass


spells = [
    Spell("Magic Missile", cost=53, damage=4),
    Spell("Drain", cost=73, damage=2, healing=2),
    Spell("Shield", cost=113, duration=6, armor=7),
    Spell("Poison", cost=173, duration=6, damage=3),
    Spell("Recharge", cost=229, duration=5, mana=101),
]
cheapest_spell_cost = min(s.cost for s in spells)
spells = {s.name: s for s in spells}


def parsed(data):
    data = {k: int(v) for k, v in (line.split(": ") for line in data.splitlines())}
    return data["Hit Points"], data["Damage"]


class GameState:

    boss_hp0, boss_damage = parsed(data)

    def __init__(
        self,
        player_hp=50,
        mana=500,
        boss_hp=boss_hp0,
        to_move="player",
        history=(),
        spent=0,
        part="a",
    ):
        self.player_hp = player_hp
        self.mana = mana
        self.boss_hp = boss_hp
        self.to_move = to_move
        self.history = history
        self.spent = spent
        self.part = part

    @property
    def winner(self):
        if self.player_hp <= 0:
            return "boss"
        if self.boss_hp <= 0:
            return "player"
        if self.to_move == "player":
            if self.mana < cheapest_spell_cost and "Recharge" not in self.active_spells:
                return "boss"
        if self.to_move == "boss":
            if "Poison" in self.active_spells:
                if self.boss_hp <= spells["Poison"].damage:
                    return "player"

    @property
    def active_spells(self):
        return {move for move, duration_left in self.history if duration_left > 0}

    def next_states(self):
        if self.winner is not None:
            return []

        player_hp = self.player_hp
        boss_hp = self.boss_hp
        mana = self.mana

        new_history = []
        for h in self.history:
            move, duration_left = h
            if duration_left > 0:
                if move == "Poison":
                    boss_hp -= spells["Poison"].damage
                elif move == "Recharge":
                    mana += spells["Recharge"].mana
                if duration_left > 1:
                    new_history.append((move, duration_left - 1))

        self.history = new_history

        if self.to_move == "boss":
            if "Shield" in self.active_spells:
                player_hp -= max(self.boss_damage - spells["Shield"].armor, 1)
            else:
                player_hp -= self.boss_damage
            return [
                GameState(
                    player_hp=player_hp,
                    mana=mana,
                    boss_hp=boss_hp,
                    to_move="player",
                    history=self.history,
                    spent=self.spent,
                    part=self.part,
                )
            ]

        assert self.to_move == "player"

        moves = []
        if self.part == "b":
            # hard mode!
            player_hp -= 1
            if player_hp <= 0:
                return moves

        for spell_name in "Shield", "Poison", "Recharge":
            spell = spells[spell_name]
            if mana - spell.cost >= 0 and spell.name not in self.active_spells:
                moves.append(
                    GameState(
                        player_hp=player_hp,
                        mana=mana - spell.cost,
                        boss_hp=boss_hp,
                        to_move="boss",
                        history=self.history + [(spell.name, spell.duration)],
                        spent=self.spent + spell.cost,
                        part=self.part,
                    )
                )

        for spell_name in "Magic Missile", "Drain":
            spell = spells[spell_name]
            if mana - spell.cost >= 0:
                moves.append(
                    GameState(
                        player_hp=player_hp + spell.healing,
                        mana=mana - spell.cost,
                        boss_hp=boss_hp - spell.damage,
                        to_move="boss",
                        history=self.history + [(spell.name, 0)],
                        spent=self.spent + spell.cost,
                        part=self.part,
                    )
                )

        return moves

    def freeze(self):
        tup = (
            self.player_hp,
            self.mana,
            self.boss_hp,
            self.to_move,
            frozenset(self.history),
            self.spent,
            self.part,
        )
        return tup

    @classmethod
    def unfreeze(cls, tup):
        return cls(*tup)

    def __eq__(self, other):
        if not isinstance(other, GameState):
            return NotImplemented
        return self.freeze() == other.freeze()

    def __hash__(self):
        return hash(self.freeze())


class Q22AStar(AStar):
    def __init__(self, part="a"):
        state0 = GameState(part=part).freeze()
        AStar.__init__(self, state0, None)

    def target_reached(self, current_state, target):
        game = GameState.unfreeze(current_state)
        return game.winner == "player"

    def adjacent(self, state):
        game = GameState.unfreeze(state)
        result = [s.freeze() for s in game.next_states()]
        return result

    def cost(self, state0, state1):
        if state1 is None:
            return 0
        game0 = GameState.unfreeze(state0)
        game1 = GameState.unfreeze(state1)
        return abs(game1.spent - game0.spent)

    heuristic = cost


astar = Q22AStar(part="a")
astar.run()
print("answer_a:", GameState.unfreeze(astar.target).spent)

astar = Q22AStar(part="b")
astar.run()
print("answer_b:", GameState.unfreeze(astar.target).spent)



"""
--- Day 9: Encoding Error ---
https://adventofcode.com/2020/day/9
"""
from aocd import data

from aoc_wim.aoc2020 import find_pair

numbers = [int(x) for x in data.splitlines()]
n = 25
if len(numbers) < n:
    n = 5  # hackish - lower "preamble" for the example test to work

pre = set(numbers[:n])
for i in range(n, len(numbers)):
    target = numbers[i]
    if find_pair(pre, target) is None:
        print("answer_a:", target)
        break
    pre.remove(numbers[i - n])
    pre.add(target)

start = 0
stop = 2
total = sum(numbers[start:stop])
while total != target or start == stop - 1:
    if total < target:
        total += numbers[stop]
        stop += 1
    else:
        total -= numbers[start]
        start += 1
chunk = numbers[start:stop]
assert sum(chunk) == target, "the chunk must sum to the target value"
assert len(chunk) >= 2, "the chunk must be a contiguous set of at least two numbers"
print("answer_b:", min(chunk) + max(chunk))



"""
--- Day 5: Doesn't He Have Intern-Elves For This? ---
https://adventofcode.com/2015/day/5
"""
from aocd import data


def vowel_count(s):
    vowels = set("aeiou")
    return sum(1 for c in s if c in vowels)


def has_double(s):
    for left, right in zip(s[:-1], s[1:]):
        if left == right:
            return True


def blacklisted(s):
    for substring in "ab", "cd", "pq", "xy":
        if substring in s:
            return True


def has_pair(s):
    for left, right in zip(s[:-1], s[1:]):
        if s.count(left + right) > 1:
            return True


def has_skip_repeat(s):
    for left, right in zip(s[:-2], s[2:]):
        if left == right:
            return True


def is_nice_a(s):
    return vowel_count(s) >= 3 and has_double(s) and not blacklisted(s)


def is_nice_b(s):
    return has_pair(s) and has_skip_repeat(s)


words = data.splitlines()
print("answer_a:", len([w for w in words if is_nice_a(w)]))
print("answer_b:", len([w for w in words if is_nice_b(w)]))



"""
--- Day 15: Dueling Generators ---
https://adventofcode.com/2017/day/15
"""
from aocd import data
from parse import parse

# Notes:
#   - The modulo (2147483647) is a mersenne prime M31: 2 ** 31 - 1
#   - The factor for generator A (16807) is 7 ** 5 (a primitive root module M31)
#   - The factor for generator B (48271) is prime
# I don't know how to exploit these facts at all, but they're actually well-known
# examples of LCGs (https://en.wikipedia.org/wiki/Linear_congruential_generator)
#   - generator A: http://www.cplusplus.com/reference/random/minstd_rand0/
#   - generator B: http://www.cplusplus.com/reference/random/minstd_rand/
# These specific factors specify a Lehmer or Park-Miller PRNG implementation:
#   https://en.wikipedia.org/wiki/Lehmer_random_number_generator
# The factors are chosen by hand because they produce a full period generating
# function with good randomness. The original 1988 paper describing this is here:
#   https://www.firstpr.com.au/dsp/rand31/p1192-park.pdf
# Maybe some obscure bit hack here can help?
#   http://graphics.stanford.edu/~seander/bithacks.html#ModulusDivision


def duel(a, b):
    fa = 16807
    fb = 48271
    d = 0x7fffffff
    count1 = 0
    a4 = []  # a values also divisible by 4
    b8 = []  # b values also divisible by 8
    for i in range(40000000):
        a = (a * fa) % d
        b = (b * fb) % d
        a16 = a & 0xffff  # lo 16 bits
        b16 = b & 0xffff
        count1 += a16 == b16
        if not a16 & 0b11:
            a4.append(a16)
        if not b16 & 0b111:
            b8.append(b16)
    # a4 will contain well over 5 million values by now, because from 40 million
    # pseudorandom numbers approx 10 million of them should be divisible by four.
    # however b8 may still be slightly under 5 million, in which case we need to
    # generate some more values.
    while len(b8) < 5000000:
        b = (b * fb) % d
        b16 = b & 0xffff
        if not b16 & 0b111:
            b8.append(b16)
    count2 = sum(a4[i] == b8[i] for i in range(5000000))
    return count1, count2


template = "Generator A starts with {:d}\nGenerator B starts with {:d}"
a0, b0 = parse(template, data).fixed
a, b = duel(a0, b0)
print("answer_a:", a)
print("answer_b:", b)
